/*
 * Filename: project_zero.c
 *
 * Description: This is the simple_central example modified to receive
 * data over BLE at a high throughput.
 *
 *
 * Copyright (C) 2015-2017 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 Generated by:
 BDS version: {{pluginInfo.AppVersion}}
 Plugin:      {{pluginInfo.Name}} {{ pluginInfo.Version}}
 Time:        {{pluginInfo.timestamp}}
 *****************************************************************************/
{% comment -%}

---------------128bit services nugget------------------------- {% endcomment -%}
{% capture advNuggetServices128b -%}
{% if gap -%}
{% unless gap.serviceAdv128b == empty -%}
{% assign advNuggetServices128bLen = gap.serviceAdv128b | size | times: 16 | plus: 2 -%}
  {{   gap.serviceAdv128b | size | times: 16 | plus: 1 }},
  GAP_ADTYPE_128BIT_MORE,
{%   for svc in gap.serviceAdv128b -%}
  {{     svc.lookup.name | upcase }}_SERV_UUID_BASE128( {{svc.lookup.name | upcase }}_SERV_UUID ),
{%   endfor -%}
{% endunless -%}
{% endif -%}
{% endcapture -%}
{% comment -%} ---------------16bit services nugget------------------------- {% endcomment -%}
{% capture advNuggetServices16b -%}
{% if gap -%}
{% unless gap.serviceAdv16b == empty -%}
{% assign advNuggetServices16bLen = gap.serviceAdv128b | size | times: 2 | plus: 2 -%}
  {{   gap.serviceAdv16b | size | times: 2 | plus: 1 }},
  GAP_ADTYPE_16BIT_MORE,
{%   for svc in gap.serviceAdv16b -%}
  LO_UINT16( {{svc.lookup.name | upcase }}_SERV_UUID ),
  HI_UINT16( {{svc.lookup.name | upcase }}_SERV_UUID ),
{%   endfor -%}
{% endunless -%}
{% endif -%}
{% endcapture -%}
{% comment -%} ---------------discovery mode define------------------------- {% endcomment -%}
{% capture advDiscMode -%}
{% if gap and gap.discMode == 'Limited' -%}
GAP_ADTYPE_FLAGS_LIMITED{% else -%}
GAP_ADTYPE_FLAGS_GENERAL{% endif -%}
{% endcapture -%}
{% comment -%} -------------device name --------------------------- {% endcomment -%}
{% capture deviceNameString -%}
{% if gap and gap.deviceName -%}
{{ gap.deviceName }}{% else -%}
Project Zero R2{% endif -%}
{% endcapture -%}
{% comment -%} -------------device name for Device Name characteristic--------------------------- {% endcomment -%}
{% capture deviceNameStringForGap -%}
{{ deviceNameString | slice: 0, 21 }}{% endcapture -%}
{% comment -%} -------------device name characters--------------------------- {% endcomment -%}
{% capture deviceNameChars -%}
'{{ deviceNameString | slice: 0, gap.useLen | split: "" | join: "', '" }}',
{% endcapture -%}
{% comment -%} -----------------Local name nugget----------------------- {% endcomment -%}
{% capture advNuggetLocalName -%}
{% assign advNuggetLocalNameLen = gap.useLen | plus: 2 -%}
  {{ gap.useLen | plus: 1 }},
  GAP_ADTYPE_LOCAL_NAME_COMPLETE,
  {{ deviceNameChars }}{% endcapture -%}
{% comment -%} ----------------Tx Power Adv------------------------{% endcomment -%}
{% capture advTxPowerNugget -%}
{% if gap -%}
  2,
  GAP_ADTYPE_POWER_LEVEL,
  {{ gap.txPower }},
{% endif -%}
{% endcapture -%}
{% comment -%} ----------------------------------------{% endcomment -%}

/*
 * INCLUDES
 */
#include <string.h>

//#define xdc_runtime_Log_DISABLE_ALL 1  // Add to disable logs from this file

#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Event.h>
#include <ti/sysbios/knl/Queue.h>
#include <ti/sysbios/knl/Clock.h>

#include <ti/drivers/PIN.h>
#include <ti/display/Display.h>

#include <xdc/runtime/Log.h>
#include <xdc/runtime/Diags.h>
#ifdef UARTLOG_ENABLE
#  include "UartLog.h"
#endif

/* This Header file contains all BLE API and icall structure definition */
#include "icall_ble_api.h"
#include <icall.h>

#include <osal_snv.h>
#include <peripheral.h>
#include <devinfoservice.h>

#include "util.h"

#include "Board.h"
#include "project_zero.h"

// Bluetooth Developer Studio services
{% for service in services -%}
#include "{{service.filename}}.h"
{% endfor -%}


/*********************************************************************
 * CONSTANTS
 */
// Advertising interval when device is discoverable (units of 625us, 160=100ms)
#define DEFAULT_ADVERTISING_INTERVAL          160

// Limited discoverable mode advertises for 30.72s, and then stops
// General discoverable mode advertises indefinitely
#define DEFAULT_DISCOVERABLE_MODE             {{ advDiscMode }}

// Default pass-code used for pairing.
#define DEFAULT_PASSCODE                      000000

// Task configuration
#define PRZ_TASK_PRIORITY                     1

#ifndef PRZ_TASK_STACK_SIZE
#define PRZ_TASK_STACK_SIZE                   800
#endif

// Internal Events for RTOS application
#define PRZ_ICALL_EVT                         ICALL_MSG_EVENT_ID  // Event_Id_31
#define PRZ_QUEUE_EVT                         UTIL_QUEUE_EVENT_ID // Event_Id_30
#define PRZ_STATE_CHANGE_EVT                  Event_Id_00
#define PRZ_CHAR_CHANGE_EVT                   Event_Id_01
#define PRZ_PERIODIC_EVT                      Event_Id_02
#define PRZ_APP_MSG_EVT                       Event_Id_03
#define PRZ_CONN_EVT_END_EVT                  Event_Id_30

#define PRZ_ALL_EVENTS                       (PRZ_ICALL_EVT        | \
                                              PRZ_QUEUE_EVT        | \
                                              PRZ_STATE_CHANGE_EVT | \
                                              PRZ_CHAR_CHANGE_EVT  | \
                                              PRZ_PERIODIC_EVT     | \
                                              PRZ_APP_MSG_EVT      | \
                                              PRZ_CONN_EVT_END_EVT)

/*********************************************************************
 * TYPEDEFS
 */
// Types of messages that can be sent to the user application task from other
// tasks or interrupts. Note: Messages from BLE Stack are sent differently.
typedef enum
{
  APP_MSG_SERVICE_WRITE = 0,   /* A characteristic value has been written     */
  APP_MSG_SERVICE_CFG,         /* A characteristic configuration has changed  */
  APP_MSG_UPDATE_CHARVAL,      /* Request from ourselves to update a value    */
  APP_MSG_GAP_STATE_CHANGE,    /* The GAP / connection state has changed      */
  APP_MSG_BUTTON_DEBOUNCED,    /* A button has been debounced with new value  */
  APP_MSG_SEND_PASSCODE,       /* A pass-code/PIN is requested during pairing */
} app_msg_types_t;

// Struct for messages sent to the application task
typedef struct
{
  Queue_Elem       _elem;
  app_msg_types_t  type;
  uint8_t          pdu[];
} app_msg_t;

// Struct for messages about characteristic data
typedef struct
{
  uint16_t svcUUID; // UUID of the service
  uint16_t dataLen; //
  uint8_t  paramID; // Index of the characteristic
  uint8_t  data[];  // Flexible array member, extended to malloc - sizeof(.)
} char_data_t;

// Struct for message about sending/requesting passcode from peer.
typedef struct
{
  uint16_t connHandle;
  uint8_t  uiInputs;
  uint8_t  uiOutputs;
  uint32   numComparison;
} passcode_req_t;

// Struct for message about button state
typedef struct
{
  PIN_Id   pinId;
  uint8_t  state;
} button_state_t;


/*********************************************************************
 * LOCAL VARIABLES
 */

// Entity ID globally used to check for source and/or destination of messages
static ICall_EntityID selfEntity;

// Sync handle globally used to post events to the application thread
static ICall_SyncHandle syncEvent;

// Queue object used for application messages.
static Queue_Struct applicationMsgQ;
static Queue_Handle hApplicationMsgQ;

{% if anyCCCDs -%}
// Clock structs for periodic notification example
{% for service in services -%}
{% for char in service.charsHaveCCCDs -%}
static Clock_Struct {{ service.nameAbbr }}_{{ char.nameCap }}_clock;
{% endfor -%}
{% endfor -%}

{% endif -%}
// Task configuration
Task_Struct przTask;
Char przTaskStack[PRZ_TASK_STACK_SIZE];


// GAP - SCAN RSP data (max size = 31 bytes)
static uint8_t scanRspData[] =
{
{% assign rspLen = 0 -%}
{% if gap.deviceNameRsp -%}
{% assign rspLen = rspLen | plus: advNuggetLocalNameLen -%}
  // complete name
{{ advNuggetLocalName }}
{% endif -%}
{% if gap and gap.servicesRsp -%}
{% assign rspLen = rspLen | plus: advNuggetServices16bLen | plus: advNuggetServices128bLen -%}
  // service UUID, to notify central devices what services are included
  // in this peripheral
{{ advNuggetServices16b }}
{{ advNuggetServices128b }}
{% endif -%}
{% if gap and gap.txPowerRsp -%}
{% assign rspLen = rspLen | plus: 3 -%}
{{advTxPowerNugget}}
{% endif -%}
{% if rspLen == 0 -%}
  // No scan response data provided.
  0x00 // Placeholder to keep the compiler happy.
{% endif -%}
};
{% if rspLen > 31 -%}

// Automatically generated warning directive.
#warning Length of Scan Response Data array ({{rspLen}}) exceeds max (31) by {{rspLen | minus: 31}} bytes.
{% endif -%}

// GAP - Advertisement data (max size = 31 bytes, though this is
// best kept short to conserve power while advertisting)
static uint8_t advertData[] =
{
  // Flags; this sets the device to use limited discoverable
  // mode (advertises for 30 seconds at a time) or general
  // discoverable mode (advertises indefinitely), depending
  // on the DEFAULT_DISCOVERY_MODE define.
  0x02,   // length of this data
  GAP_ADTYPE_FLAGS,
  DEFAULT_DISCOVERABLE_MODE | GAP_ADTYPE_FLAGS_BREDR_NOT_SUPPORTED,
{% assign rspLen = 3 -%}
{% if gap.deviceNameAdv -%}
{% assign rspLen = rspLen | plus: advNuggetLocalNameLen -%}

  // complete name
{{ advNuggetLocalName }}
{% endif -%}
{% if gap and gap.servicesAdv -%}
{% assign rspLen = rspLen | plus: advNuggetServices16bLen | plus: advNuggetServices128bLen -%}

  // service UUID, to notify central devices what services are included
  // in this peripheral
{{ advNuggetServices16b }}
{{ advNuggetServices128b }}
{% endif -%}
{% if gap and gap.txPowerAdv -%}
{% assign rspLen = rspLen | plus: 3 -%}
{{advTxPowerNugget}}
{% endif -%}
};
{% if rspLen > 31 -%}

// Automatically generated warning directive.
#warning Length of Advertisement Data array ({{rspLen}}) exceeds max (31) by {{rspLen | minus: 31}} bytes.
{% endif -%}

// GAP GATT Attributes
static uint8_t attDeviceName[GAP_DEVICE_NAME_LEN] = "{{ deviceNameStringForGap }}";

// Globals used for ATT Response retransmission
static gattMsgEvent_t *pAttRsp = NULL;
static uint8_t rspTxRetry = 0;
{% if pins.numPins -%}

static PIN_Handle userPinHandle;
static PIN_State  userPinState;

PIN_Config userPinTable[] = {
{%   for pin in pins.list -%}
{%     if pin.input or pin.button -%}
  {{ pin.define }} | PIN_INPUT_EN | PIN_PULLUP{% if pin.irq %} | PIN_IRQ_NEGEDGE{% endif %},
{%   endif -%}
{%     if pin.output -%}
  {{ pin.define }} | PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
{%     endif -%}
{%   endfor -%}
  PIN_TERMINATE
};

{%   for pin in pins.list -%}
{%     if pin.button -%}
static Clock_Struct {{pin.nameCap}}DebounceClock;
static uint8_t {{pin.nameCap}}State = 0;

{%     endif -%}
{%   endfor -%}
{% endif -%}

// Global display handle
Display_Handle dispHandle;


/*********************************************************************
 * LOCAL FUNCTIONS
 */

static void ProjectZero_init( void );
static void ProjectZero_taskFxn(UArg a0, UArg a1);

static void user_processApplicationMessage(app_msg_t *pMsg);
static uint8_t ProjectZero_processStackMsg(ICall_Hdr *pMsg);
static uint8_t ProjectZero_processGATTMsg(gattMsgEvent_t *pMsg);

static void ProjectZero_sendAttRsp(void);
static uint8_t ProjectZero_processGATTMsg(gattMsgEvent_t *pMsg);
static void ProjectZero_freeAttRsp(uint8_t status);

static void user_processGapStateChangeEvt(gaprole_States_t newState);
static void user_gapStateChangeCB(gaprole_States_t newState);
static void user_gapBondMgr_passcodeCB(uint8_t *deviceAddr, uint16_t connHandle,
                                       uint8_t uiInputs, uint8_t uiOutputs, uint32 numComparison);
static void user_gapBondMgr_pairStateCB(uint16_t connHandle, uint8_t state,
                                        uint8_t status);

{% if pins.numButtons or pins.numInput -%}
static void pinHwiFxn(PIN_Handle handle, PIN_Id pinId);
{% endif -%}
{% if pins.numButtons -%}
static void buttonDebounceSwiFxn(UArg buttonId);
static void user_handleButtonPress(button_state_t *pState);
{% endif -%}

// Generic callback handlers for value changes in services.
static void user_service_ValueChangeCB( uint16_t connHandle, uint16_t svcUuid, uint8_t paramID, uint8_t *pValue, uint16_t len );
static void user_service_CfgChangeCB( uint16_t connHandle, uint16_t svcUuid, uint8_t paramID, uint8_t *pValue, uint16_t len );

{% if anyCCCDs or anyWriteable -%}
// Task context handlers for generated services.
{% endif -%}
{% for service in services -%}
{% if service.writeable -%}
static void user_{{service.nameCap}}_ValueChangeHandler(char_data_t *pCharData);
{% endif -%}
{% if service.numCCCDs -%}
static void user_{{service.nameCap}}_CfgChangeHandler(char_data_t *pCharData);
{% endif -%}

{% endfor -%}
{% if anyCCCDs -%}
// Callback handler(s) for the clock object(s) used to demonstrate notifications
{% for service in services -%}
{%   if service.numCCCDs and service.numCCCDs > service.pins.numInput -%}
static void user_{{service.nameCap}}_clockSwiHandler(UArg paramID);
{%   endif -%}
{% endfor -%}

// Task handler for sending notifications.
static void user_updateCharVal(char_data_t *pCharData);
{% endif -%}

// Utility functions
static void user_enqueueRawAppMsg(app_msg_types_t appMsgType, uint8_t *pData, uint16_t len );
{% if anyWriteable or anyCCCDs -%}
static void user_enqueueCharDataMsg( app_msg_types_t appMsgType, uint16_t connHandle,
                                       uint16_t serviceUUID, uint8_t paramID,
                                       uint8_t *pValue, uint16_t len );

{% endif -%}
static char *Util_getLocalNameStr(const uint8_t *data);
static char *Util_convertArrayToHexString(uint8_t const *src, uint8_t src_len,
                                          uint8_t *dst, uint8_t dst_len);

/*********************************************************************
 * PROFILE CALLBACKS
 */

// GAP Role Callbacks
static gapRolesCBs_t user_gapRoleCBs =
{
  user_gapStateChangeCB     // Profile State Change Callbacks
};

// GAP Bond Manager Callbacks
static gapBondCBs_t user_bondMgrCBs =
{
  user_gapBondMgr_passcodeCB, // Passcode callback
  user_gapBondMgr_pairStateCB // Pairing / Bonding state Callback
};

/*
 * Callbacks in the user application for events originating from BLE services.
 */
// Service callback structure for registering our handlers with the service
{% for service in services -%}
{% if service.numCCCDs or service.numWriteable -%}
// {{ service.nameNice }} callback handler.
// The type {{service.name}}CBs_t is defined in {{service.name}}.h
static {{service.nameCap}}CBs_t user_{{service.name}}CBs =
{
{% if service.writeable -%}
  .pfnChangeCb    = user_service_ValueChangeCB, // Characteristic value change callback handler
{% else -%}
  .pfnChangeCb    = NULL, // No writable chars in {{ service.nameNice }}, so no change handler.
{% endif -%}
{% if service.numCCCDs -%}
  .pfnCfgChangeCb = user_service_CfgChangeCB, // Noti/ind configuration callback handler
{% else -%}
  .pfnCfgChangeCb = NULL, // No notification-/indication enabled chars in {{ service.nameNice }}
{% endif -%}
};

{% endif -%}
{% endfor -%}


/*********************************************************************
 * PUBLIC FUNCTIONS
 */

/*
 * @brief   Task creation function for the user task.
 *
 * @param   None.
 *
 * @return  None.
 */
void ProjectZero_createTask(void)
{
  Task_Params taskParams;

  // Configure task
  Task_Params_init(&taskParams);
  taskParams.stack = przTaskStack;
  taskParams.stackSize = PRZ_TASK_STACK_SIZE;
  taskParams.priority = PRZ_TASK_PRIORITY;

  Task_construct(&przTask, ProjectZero_taskFxn, &taskParams, NULL);
}

/*
 * @brief   Called before the task loop and contains application-specific
 *          initialization of the BLE stack, hardware setup, power-state
 *          notification if used, and BLE profile/service initialization.
 *
 * @param   None.
 *
 * @return  None.
 */
static void ProjectZero_init(void)
{
  // ******************************************************************
  // NO STACK API CALLS CAN OCCUR BEFORE THIS CALL TO ICall_registerApp
  // ******************************************************************
  // Register the current thread as an ICall dispatcher application
  // so that the application can send and receive messages via ICall to Stack.
  ICall_registerApp(&selfEntity, &syncEvent);

  Log_info0("Initializing the user task, hardware, BLE stack and services.");

  // Open display. By default this is disabled via the predefined symbol Display_DISABLE_ALL.
  dispHandle = Display_open(Display_Type_LCD, NULL);

  // Initialize queue for application messages.
  // Note: Used to transfer control to application thread from e.g. interrupts.
  Queue_construct(&applicationMsgQ, NULL);
  hApplicationMsgQ = Queue_handle(&applicationMsgQ);

  // ******************************************************************
  // Hardware initialization
  // ******************************************************************

{% if pins.numPins -%}
  // Open user PINs
  userPinHandle = PIN_open(&userPinState, userPinTable);
  if(!userPinHandle) {
    Log_error0("Error initializing board pins");
    Task_exit();
  }
{%   if pins.numButtons -%}

  // Setup callback for button pins
  if (PIN_registerIntCb(userPinHandle, &pinHwiFxn) != 0) {
    Log_error0("Error registering button callback function");
    Task_exit();
  }

  // Create the debounce clock objects
  Clock_Params clockParams;
  Clock_Params_init(&clockParams);

  // All debounce clocks use the same callback, so differentiate on argument
  // given to the callback in Swi context
{%     for pin in pins.list -%}
{%       if pin.button -%}
  clockParams.arg = {{ pin.define }};

  // Initialize to 50 ms timeout when Clock_start is called.
  // Timeout argument is in ticks, so convert from ms to ticks via tickPeriod.
  Clock_construct(&{{ pin.nameCap }}DebounceClock, buttonDebounceSwiFxn,
                  50 * (1000/Clock_tickPeriod),
                  &clockParams);

{%       endif -%}
{%     endfor -%}
{%   endif -%}
{% endif -%}

{% if anyCCCDs -%}
{% if numCCCDs > pins.numInput -%}
  // ******************************************************************
  // Initialization of clock objects used for notifiable characterisics
  // ******************************************************************
{%   unless pins.numButtons -%}
  Clock_Params clockParams;
  Clock_Params_init(&clockParams);
{%   endunless -%}
  clockParams.period = 5000 * (1000 / Clock_tickPeriod); // 5000ms period

  // Clock struct initialization for periodic notification example
  // Clock callbacks only have one parameter, so make one callback handler per service
  // and one Clock Struct per noti/ind characteristic.
{% for service in services -%}
{% for char in service.charsHaveCCCDs -%}
{% unless char.pin.button or char.pin.input -%}
  clockParams.arg = {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_ID;
  Clock_construct(&{{ service.nameAbbr }}_{{ char.nameCap }}_clock,
                  user_{{service.nameCap}}_clockSwiHandler,
                  0,
                  &clockParams);
{% endunless -%}
{% endfor -%}
{% endfor -%}

{% endif -%}
{% endif -%}
  // ******************************************************************
  // BLE Stack initialization
  // ******************************************************************

  // Setup the GAP Peripheral Role Profile
  uint8_t initialAdvertEnable = TRUE;  // Advertise on power-up

  // By setting this to zero, the device will go into the waiting state after
  // being discoverable. Otherwise wait this long [ms] before advertising again.
  uint16_t advertOffTime = 0; // miliseconds

  // Set advertisement enabled.
  GAPRole_SetParameter(GAPROLE_ADVERT_ENABLED, sizeof(uint8_t),
                       &initialAdvertEnable);

  // Configure the wait-time before restarting advertisement automatically
  GAPRole_SetParameter(GAPROLE_ADVERT_OFF_TIME, sizeof(uint16_t),
                       &advertOffTime);

  // Initialize Scan Response data
  GAPRole_SetParameter(GAPROLE_SCAN_RSP_DATA, sizeof(scanRspData), scanRspData);

  // Initialize Advertisement data
  GAPRole_SetParameter(GAPROLE_ADVERT_DATA, sizeof(advertData), advertData);

  Log_info1("Name in advertData array: \x1b[33m%s\x1b[0m",
            (IArg)Util_getLocalNameStr(advertData));

  // Set advertising interval
  uint16_t advInt = DEFAULT_ADVERTISING_INTERVAL;

  GAP_SetParamValue(TGAP_LIM_DISC_ADV_INT_MIN, advInt);
  GAP_SetParamValue(TGAP_LIM_DISC_ADV_INT_MAX, advInt);
  GAP_SetParamValue(TGAP_GEN_DISC_ADV_INT_MIN, advInt);
  GAP_SetParamValue(TGAP_GEN_DISC_ADV_INT_MAX, advInt);

  // Set duration of advertisement before stopping in Limited adv mode.
  GAP_SetParamValue(TGAP_LIM_ADV_TIMEOUT, 30); // Seconds

  // ******************************************************************
  // BLE Bond Manager initialization
  // ******************************************************************
  uint32_t passkey = 0; // passkey "000000"
  uint8_t pairMode = GAPBOND_PAIRING_MODE_WAIT_FOR_REQ;
  uint8_t mitm = TRUE;
  uint8_t ioCap = GAPBOND_IO_CAP_DISPLAY_ONLY;
  uint8_t bonding = TRUE;

  GAPBondMgr_SetParameter(GAPBOND_DEFAULT_PASSCODE, sizeof(uint32_t),
                          &passkey);
  GAPBondMgr_SetParameter(GAPBOND_PAIRING_MODE, sizeof(uint8_t), &pairMode);
  GAPBondMgr_SetParameter(GAPBOND_MITM_PROTECTION, sizeof(uint8_t), &mitm);
  GAPBondMgr_SetParameter(GAPBOND_IO_CAPABILITIES, sizeof(uint8_t), &ioCap);
  GAPBondMgr_SetParameter(GAPBOND_BONDING_ENABLED, sizeof(uint8_t), &bonding);

  // ******************************************************************
  // BLE Service initialization
  // ******************************************************************

  // Add services to GATT server
  GGS_AddService(GATT_ALL_SERVICES);           // GAP
  GATTServApp_AddService(GATT_ALL_SERVICES);   // GATT attributes
  DevInfo_AddService();                        // Device Information Service

  // Set the device name characteristic in the GAP Profile
  GGS_SetParameter(GGS_DEVICE_NAME_ATT, GAP_DEVICE_NAME_LEN, attDeviceName);

// Add services to GATT server and give ID of this task for Indication acks.
{% for service in services -%}
  {{ service.nameCap }}_AddService( selfEntity );
{% endfor -%}

{% if anyWriteable or anyCCCDs -%}
  // Register callbacks with the generated services that
  // can generate events (writes received) to the application
{% for service in services -%}
{% if service.writeable or service.numCCCDs -%}
  {{ service.nameCap }}_RegisterAppCBs( &user_{{service.name}}CBs );
{% endif -%}
{% endfor -%}
{% endif -%}
{% if anyReadable or anyCCCDs -%}
  // Placeholder variable for characteristic intialization
  uint8_t someVal[20] = {0};

{% endif -%}
{% for service in services -%}
{% if service.numCCCDs or service.numReadable -%}
  // Initalization of characteristics in {{service.name}} that can provide data
  // to a peer device over the air.
{% for char in service.charsNeedRead -%}
  {{ service.nameCap }}_SetParameter({{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID, {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_LEN, &someVal);
{% endfor -%}

{% endif -%}
{% endfor -%}


  // Start the stack in Peripheral mode.
  VOID GAPRole_StartDevice(&user_gapRoleCBs);

  // Start Bond Manager
  VOID GAPBondMgr_Register(&user_bondMgrCBs);

  // Register with GAP for HCI/Host messages
  GAP_RegisterForMsgs(selfEntity);

  // Register for GATT local events and ATT Responses pending for transmission
  GATT_RegisterForMsgs(selfEntity);
}


/*
 * @brief   Application task entry point.
 *
 *          Invoked by TI-RTOS when BIOS_start is called. Calls an init function
 *          and enters an infinite loop waiting for messages.
 *
 *          Messages can be either directly from the BLE stack or from user code
 *          like Hardware Interrupt (Hwi) or a callback function.
 *
 *          The reason for sending messages to this task from e.g. Hwi's is that
 *          some RTOS and Stack APIs are not available in callbacks and so the
 *          actions that may need to be taken is dispatched to this Task.
 *
 * @param   a0, a1 - not used.
 *
 * @return  None.
 */
static void ProjectZero_taskFxn(UArg a0, UArg a1)
{
  // Initialize application
  ProjectZero_init();

  // Application main loop
  for (;;)
  {
    uint32_t events;

    // Waits for an event to be posted associated with the calling thread.
    // Note that an event associated with a thread is posted when a
    // message is queued to the message receive queue of the thread
    events = Event_pend(syncEvent, Event_Id_NONE, PRZ_ALL_EVENTS,
                        ICALL_TIMEOUT_FOREVER);

    if (events)
    {
      ICall_EntityID dest;
      ICall_ServiceEnum src;
      ICall_HciExtEvt *pMsg = NULL;

      // Check if we got a signal because of a stack message
      if (ICall_fetchServiceMsg(&src, &dest,
                                (void **)&pMsg) == ICALL_ERRNO_SUCCESS)
      {
        uint8 safeToDealloc = TRUE;

        if ((src == ICALL_SERVICE_CLASS_BLE) && (dest == selfEntity))
        {
          ICall_Stack_Event *pEvt = (ICall_Stack_Event *)pMsg;

          // Check for event flags received (event signature 0xffff)
          if (pEvt->signature == 0xffff)
          {
            // Event received when a connection event is completed
            if (pEvt->event_flag & PRZ_CONN_EVT_END_EVT)
            {
              // Try to retransmit pending ATT Response (if any)
              ProjectZero_sendAttRsp();
            }
          }
          else // It's a message from the stack and not an event.
          {
            // Process inter-task message
            safeToDealloc = ProjectZero_processStackMsg((ICall_Hdr *)pMsg);
          }
        }

        if (pMsg && safeToDealloc)
        {
          ICall_freeMsg(pMsg);
        }
      }

      // Process messages sent from another task or another context.
      while (!Queue_empty(hApplicationMsgQ))
      {
        app_msg_t *pMsg = Queue_dequeue(hApplicationMsgQ);

        // Process application-layer message probably sent from ourselves.
        user_processApplicationMessage(pMsg);

        // Free the received message.
        ICall_free(pMsg);
      }

    }
  }
}


/*
 * @brief   Handle application messages
 *
 *          These are messages not from the BLE stack, but from the
 *          application itself.
 *
 *          For example, in a Software Interrupt (Swi) it is not possible to
 *          call any BLE APIs, so instead the Swi function must send a message
 *          to the application Task for processing in Task context.
 *
 * @param   pMsg  Pointer to the message of type app_msg_t.
 *
 * @return  None.
 */
static void user_processApplicationMessage(app_msg_t *pMsg)
{
{% if anyWriteable or anyCCCDs -%}
        char_data_t *pCharData = (char_data_t *)pMsg->pdu;

{% endif -%}

  switch (pMsg->type)
  {
{% if anyWriteable -%}
          case APP_MSG_SERVICE_WRITE: /* Message about received value write */
            /* Call different handler per service */
            switch(pCharData->svcUUID) {
{%   for service in services -%}
{%     if service.writeable -%}
              case {{ service.name | upcase }}_SERV_UUID:
                user_{{service.nameCap}}_ValueChangeHandler(pCharData);
                break;
{%     endif -%}
{%   endfor -%}
            }
            break;
{% endif -%}
{% if anyCCCDs -%}

          case APP_MSG_SERVICE_CFG: /* Message about received CCCD write */
            /* Call different handler per service */
            switch(pCharData->svcUUID) {
{%   for service in services -%}
{%     if service.numCCCDs -%}
              case {{ service.name | upcase }}_SERV_UUID:
                user_{{service.nameCap}}_CfgChangeHandler(pCharData);
                break;
{%     endif -%}
{%   endfor -%}
            }
            break;

          case APP_MSG_UPDATE_CHARVAL: /* Message to self from to update a value */
            user_updateCharVal(pCharData);
            break;

{% endif -%}

    case APP_MSG_GAP_STATE_CHANGE: /* Message that GAP state changed  */
      user_processGapStateChangeEvt( *(gaprole_States_t *)pMsg->pdu );
      break;

    case APP_MSG_SEND_PASSCODE: /* Message about pairing PIN request */
      {
        passcode_req_t *pReq = (passcode_req_t *)pMsg->pdu;
        Log_info2("BondMgr Requested passcode. We are %s passcode %06d",
                  (IArg)(pReq->uiInputs?"Sending":"Displaying"),
                  DEFAULT_PASSCODE);
        // Send passcode response.
        GAPBondMgr_PasscodeRsp(pReq->connHandle, SUCCESS, DEFAULT_PASSCODE);
      }
      break;
{% if pins.numButtons -%}
    case APP_MSG_BUTTON_DEBOUNCED: /* Message from swi about pin change */
      {
        button_state_t *pButtonState = (button_state_t *)pMsg->pdu;
        user_handleButtonPress(pButtonState);
      }
      break;
{% endif -%}
  }
}


/******************************************************************************
 *****************************************************************************
 *
 *  Handlers of system/application events deferred to the user Task context.
 *  Invoked from the application Task function above.
 *
 *  Further down you can find the callback handler section containing the
 *  functions that defer their actions via messages to the application task.
 *
 ****************************************************************************
 *****************************************************************************/


/*
 * @brief   Process a pending GAP Role state change event.
 *
 * @param   newState - new state
 *
 * @return  None.
 */
static void user_processGapStateChangeEvt(gaprole_States_t newState)
{
  switch ( newState )
  {
    case GAPROLE_STARTED:
      {
        uint8_t ownAddress[B_ADDR_LEN];
        uint8_t systemId[DEVINFO_SYSTEM_ID_LEN];

        GAPRole_GetParameter(GAPROLE_BD_ADDR, ownAddress);

        // use 6 bytes of device address for 8 bytes of system ID value
        systemId[0] = ownAddress[0];
        systemId[1] = ownAddress[1];
        systemId[2] = ownAddress[2];

        // set middle bytes to zero
        systemId[4] = 0x00;
        systemId[3] = 0x00;

        // shift three bytes up
        systemId[7] = ownAddress[5];
        systemId[6] = ownAddress[4];
        systemId[5] = ownAddress[3];

        DevInfo_SetParameter(DEVINFO_SYSTEM_ID, DEVINFO_SYSTEM_ID_LEN, systemId);

        // Display device address
        char *cstr_ownAddress = Util_convertBdAddr2Str(ownAddress);
        Log_info1("GAP is started. Our address: \x1b[32m%s\x1b[0m", (IArg)cstr_ownAddress);
      }
      break;

    case GAPROLE_ADVERTISING:
      Log_info0("Advertising");
      break;

    case GAPROLE_CONNECTED:
      {
        uint8_t peerAddress[B_ADDR_LEN];

        GAPRole_GetParameter(GAPROLE_CONN_BD_ADDR, peerAddress);

        char *cstr_peerAddress = Util_convertBdAddr2Str(peerAddress);
        Log_info1("Connected. Peer address: \x1b[32m%s\x1b[0m", (IArg)cstr_peerAddress);
       }
      break;

    case GAPROLE_CONNECTED_ADV:
      Log_info0("Connected and advertising");
      break;

    case GAPROLE_WAITING:
{% if anyCCCDs -%}
{% if numCCCDs > pins.numButtons -%}

      // Turn off periodic clocks for ind/noti demo
{%   for service in services -%}
{%     for char in service.charsHaveCCCDs -%}
      Clock_stop((Clock_Handle)&{{ service.nameAbbr }}_{{ char.nameCap }}_clock);
{%     endfor -%}
{%   endfor -%}
{% endif -%}
{% endif -%}
      Log_info0("Disconnected / Idle");
      break;

    case GAPROLE_WAITING_AFTER_TIMEOUT:
{% if anyCCCDs -%}
{% if numCCCDs > pins.numButtons -%}

      // Turn off periodic clocks for ind/noti demo
{%   for service in services -%}
{%     for char in service.charsHaveCCCDs -%}
      Clock_stop((Clock_Handle)&{{ service.nameAbbr }}_{{ char.nameCap }}_clock);
{%     endfor -%}
{%   endfor -%}
{% endif -%}
{% endif -%}
      Log_info0("Connection timed out");
      break;

    case GAPROLE_ERROR:
      Log_info0("Error");
      break;

    default:
      break;
  }
}


{% if pins.numButtons -%}
/*
 * @brief   Handle a debounced button press or release in Task context.
 *          Invoked by the taskFxn based on a message received from a callback.
 *
 * @see     buttonDebounceSwiFxn
 * @see     pinHwiFxn
 *
 * @param   pState  pointer to button_state_t message sent from debounce Swi.
 *
 * @return  None.
 */
static void user_handleButtonPress(button_state_t *pState)
{
  Log_info2("Button mapped to IOID_%d %s",
    (IArg)pState->pinId,
    (IArg)(pState->state?"\x1b[32mpressed\x1b[0m":
                         "\x1b[33mreleased\x1b[0m"));

  // Update the service with the new value.
  // Will automatically send notification/indication if enabled.
  switch (pState->pinId)
  {
{%   for pin in pins.list -%}
{%     if pin.button -%}
    case {{ pin.define }}:
      {{ pin.service.nameCap }}_SetParameter({{ pin.service.nameAbbr | upcase }}_{{ pin.char.name | upcase }}_ID,
                                 sizeof(pState->state),
                                 &pState->state);
      break;
{%     endif -%}
{%   endfor -%}
  }
}
{% endif -%}

{% for service in services -%}
{% if service.writeable -%}

/*
 * @brief   Handle a write request sent from a peer device.
 *
 *          Invoked by the Task based on a message received from a callback.
 *
 *          When we get here, the request has already been accepted by the
 *          service and is valid from a BLE protocol perspective as well as
 *          having the correct length as defined in the service implementation.
 *
 * @param   pCharData  pointer to malloc'd char write data
 *
 * @return  None.
 */
void user_{{service.nameCap}}_ValueChangeHandler(char_data_t *pCharData)
{
  static uint8_t pretty_data_holder[16]; // 5 bytes as hex string "AA:BB:CC:DD:EE"
  Util_convertArrayToHexString(pCharData->data, pCharData->dataLen,
                               pretty_data_holder, sizeof(pretty_data_holder));

  switch (pCharData->paramID)
  {
{% for char in service.charsNeedWrite -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
      Log_info3("Value Change msg: %s %s: %s",
                (IArg)"{{service.nameNice}}",
                (IArg)"{{char.nameNice}}",
                (IArg)pretty_data_holder);

      // Do something useful with pCharData->data here
      // -------------------------
{%   if char.pin.output -%}
      // Set the output value equal to the received value. 0 is off, not 0 is on
      PIN_setOutputValue(userPinHandle, {{char.pin.define}}, pCharData->data[0]);
      Log_info2("Setting %s %s",
                (IArg)"{{char.pin.define}}",
                (IArg)(pCharData->data[0]?"high":"low"));
{%   endif -%}
      break;

{% endfor -%}
  default:
    return;
  }
}
{% endif -%}
{%   if service.numCCCDs -%}

/*
 * @brief   Handle a CCCD (configuration change) write received from a peer
 *          device. This tells us whether the peer device wants us to send
 *          Notifications or Indications.
 *
 * @param   pCharData  pointer to malloc'd char write data
 *
 * @return  None.
 */
void user_{{service.nameCap}}_CfgChangeHandler(char_data_t *pCharData)
{
  // Cast received data to uint16, as that's the format for CCCD writes.
  uint16_t configValue = *(uint16_t *)pCharData->data;
  char *configValString;

  // Determine what to tell the user
  switch(configValue)
  {
  case GATT_CFG_NO_OPERATION:
    configValString = "Noti/Ind disabled";
    break;
  case GATT_CLIENT_CFG_NOTIFY:
    configValString = "Notifications enabled";
    break;
  case GATT_CLIENT_CFG_INDICATE:
    configValString = "Indications enabled";
    break;
  }

  switch (pCharData->paramID)
  {
{%     for char in service.charsHaveCCCDs -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
      Log_info3("CCCD Change msg: %s %s: %s",
                (IArg)"{{service.nameNice}}",
                (IArg)"{{char.nameNice}}",
                (IArg)configValString);
      // -------------------------
      // Do something useful with configValue here. It tells you whether someone
      // wants to know the state of this characteristic.
{%       unless char.pin.input or char.pin.button -%}
      // ... In the generated example we turn periodic clocks on/off
      if (configValue) // 0x0001 and 0x0002 both indicate turned on.
        Clock_start((Clock_Handle)&{{ service.nameAbbr }}_{{ char.nameCap }}_clock);
      else
        Clock_stop((Clock_Handle)&{{ service.nameAbbr }}_{{ char.nameCap }}_clock);
{%       endunless -%}
      break;

{%     endfor -%}
  default:
    return;
  }
}
{%   endif -%}
{% endfor -%}


/*
 * @brief   Process an incoming BLE stack message.
 *
 *          This could be a GATT message from a peer device like acknowledgement
 *          of an Indication we sent, or it could be a response from the stack
 *          to an HCI message that the user application sent.
 *
 * @param   pMsg - message to process
 *
 * @return  TRUE if safe to deallocate incoming message, FALSE otherwise.
 */
static uint8_t ProjectZero_processStackMsg(ICall_Hdr *pMsg)
{
  uint8_t safeToDealloc = TRUE;

  switch (pMsg->event)
  {
    case GATT_MSG_EVENT:
      // Process GATT message
      safeToDealloc = ProjectZero_processGATTMsg((gattMsgEvent_t *)pMsg);
      break;

    case HCI_GAP_EVENT_EVENT:
      {
        // Process HCI message
        switch(pMsg->status)
        {
          case HCI_COMMAND_COMPLETE_EVENT_CODE:
            // Process HCI Command Complete Event
            Log_info0("HCI Command Complete Event received");
            break;

          default:
            break;
        }
      }
      break;

    default:
      // do nothing
      break;
  }

  return (safeToDealloc);
}


/*
 * @brief   Process GATT messages and events.
 *
 * @return  TRUE if safe to deallocate incoming message, FALSE otherwise.
 */
static uint8_t ProjectZero_processGATTMsg(gattMsgEvent_t *pMsg)
{
  // See if GATT server was unable to transmit an ATT response
  if (pMsg->hdr.status == blePending)
  {
    Log_warning1("Outgoing RF FIFO full. Re-schedule transmission of msg with opcode 0x%02x",
      pMsg->method);

    // No HCI buffer was available. Let's try to retransmit the response
    // on the next connection event.
    if (HCI_EXT_ConnEventNoticeCmd(pMsg->connHandle, selfEntity,
                                   PRZ_CONN_EVT_END_EVT) == SUCCESS)
    {
      // First free any pending response
      ProjectZero_freeAttRsp(FAILURE);

      // Hold on to the response message for retransmission
      pAttRsp = pMsg;

      // Don't free the response message yet
      return (FALSE);
    }
  }
  else if (pMsg->method == ATT_FLOW_CTRL_VIOLATED_EVENT)
  {
    // ATT request-response or indication-confirmation flow control is
    // violated. All subsequent ATT requests or indications will be dropped.
    // The app is informed in case it wants to drop the connection.

    // Log the opcode of the message that caused the violation.
    Log_error1("Flow control violated. Opcode of offending ATT msg: 0x%02x",
      pMsg->msg.flowCtrlEvt.opcode);
  }
  else if (pMsg->method == ATT_MTU_UPDATED_EVENT)
  {
    // MTU size updated
    Log_info1("MTU Size change: %d bytes", pMsg->msg.mtuEvt.MTU);
  }
  else
  {
    // Got an expected GATT message from a peer.
    Log_info1("Recevied GATT Message. Opcode: 0x%02x", pMsg->method);
  }

  // Free message payload. Needed only for ATT Protocol messages
  GATT_bm_free(&pMsg->msg, pMsg->method);

  // It's safe to free the incoming message
  return (TRUE);
}




/*
 *  Application error handling functions
 *****************************************************************************/

/*
 * @brief   Send a pending ATT response message.
 *
 *          The message is one that the stack was trying to send based on a
 *          peer request, but the response couldn't be sent because the
 *          user application had filled the TX queue with other data.
 *
 * @param   none
 *
 * @return  none
 */
static void ProjectZero_sendAttRsp(void)
{
  // See if there's a pending ATT Response to be transmitted
  if (pAttRsp != NULL)
  {
    uint8_t status;

    // Increment retransmission count
    rspTxRetry++;

    // Try to retransmit ATT response till either we're successful or
    // the ATT Client times out (after 30s) and drops the connection.
    status = GATT_SendRsp(pAttRsp->connHandle, pAttRsp->method, &(pAttRsp->msg));
    if ((status != blePending) && (status != MSG_BUFFER_NOT_AVAIL))
    {
      // Disable connection event end notice
      HCI_EXT_ConnEventNoticeCmd(pAttRsp->connHandle, selfEntity, 0);

      // We're done with the response message
      ProjectZero_freeAttRsp(status);
    }
    else
    {
      // Continue retrying
      Log_warning2("Retrying message with opcode 0x%02x. Attempt %d",
        pAttRsp->method, rspTxRetry);
    }
  }
}

/*
 * @brief   Free ATT response message.
 *
 * @param   status - response transmit status
 *
 * @return  none
 */
static void ProjectZero_freeAttRsp(uint8_t status)
{
  // See if there's a pending ATT response message
  if (pAttRsp != NULL)
  {
    // See if the response was sent out successfully
    if (status == SUCCESS)
    {
      Log_info2("Sent message with opcode 0x%02x. Attempt %d",
        pAttRsp->method, rspTxRetry);
    }
    else
    {
      Log_error2("Gave up message with opcode 0x%02x. Status: %d",
        pAttRsp->method, status);

      // Free response payload
      GATT_bm_free(&pAttRsp->msg, pAttRsp->method);
    }

    // Free response message
    ICall_freeMsg(pAttRsp);

    // Reset our globals
    pAttRsp = NULL;
    rspTxRetry = 0;
  }
}


/******************************************************************************
 *****************************************************************************
 *
 *  Handlers of direct system callbacks.
 *
 *  Typically enqueue the information or request as a message for the
 *  application Task for handling.
 *
 ****************************************************************************
 *****************************************************************************/


/*
 *  Callbacks from the Stack Task context (GAP or Service changes)
 *****************************************************************************/

/**
 * Callback from GAP Role indicating a role state change.
 */
static void user_gapStateChangeCB(gaprole_States_t newState)
{
  Log_info1("(CB) GAP State change: %d, Sending msg to app.", (IArg)newState);
  user_enqueueRawAppMsg( APP_MSG_GAP_STATE_CHANGE, (uint8_t *)&newState, sizeof(newState) );
}

/*
 * @brief   Passcode callback.
 *
 * @param   connHandle - connection handle
 * @param   uiInputs   - input passcode?
 * @param   uiOutputs  - display passcode?
 * @param   numComparison - numeric comparison value
 *
 * @return  none
 */
static void user_gapBondMgr_passcodeCB(uint8_t *deviceAddr, uint16_t connHandle,
                                       uint8_t uiInputs, uint8_t uiOutputs, uint32 numComparison)
{
  passcode_req_t req =
  {
    .connHandle = connHandle,
    .uiInputs = uiInputs,
    .uiOutputs = uiOutputs,
    .numComparison = numComparison
  };

  // Defer handling of the passcode request to the application, in case
  // user input is required, and because a BLE API must be used from Task.
  user_enqueueRawAppMsg(APP_MSG_SEND_PASSCODE, (uint8_t *)&req, sizeof(req));
}

/*
 * @brief   Pairing state callback.
 *
 * @param   connHandle - connection handle
 * @param   state      - pairing state
 * @param   status     - pairing status
 *
 * @return  none
 */
static void user_gapBondMgr_pairStateCB(uint16_t connHandle, uint8_t state,
                                        uint8_t status)
{
  if (state == GAPBOND_PAIRING_STATE_STARTED)
  {
    Log_info0("Pairing started");
  }
  else if (state == GAPBOND_PAIRING_STATE_COMPLETE)
  {
    if (status == SUCCESS)
    {
      Log_info0("Pairing completed successfully.");
    }
    else
    {
      Log_error1("Pairing failed. Error: %02x", status);
    }
  }
  else if (state == GAPBOND_PAIRING_STATE_BONDED)
  {
    if (status == SUCCESS)
    {
     Log_info0("Re-established pairing from stored bond info.");
    }
  }
}

{% if anyWriteable -%}
/**
 * Callback handler for characteristic value changes in services.
 */
static void user_service_ValueChangeCB( uint16_t connHandle, uint16_t svcUuid,
                                        uint8_t paramID, uint8_t *pValue,
                                        uint16_t len )
{
  // See the service header file to compare paramID with characteristic.
  Log_info2("(CB) Characteristic value change: svc(0x%04x) paramID(%d). "
            "Sending msg to app.", (IArg)svcUuid, (IArg)paramID);
  user_enqueueCharDataMsg(APP_MSG_SERVICE_WRITE, connHandle, svcUuid, paramID,
                          pValue, len);
}

{% endif -%}
{% if anyCCCDs -%}
/**
 * Callback handler for characteristic configuration changes in services.
 */
static void user_service_CfgChangeCB( uint16_t connHandle, uint16_t svcUuid,
                                      uint8_t paramID, uint8_t *pValue,
                                      uint16_t len )
{
  Log_info2("(CB) Char config change: svc(0x%04x) paramID(%d). "
            "Sending msg to app.", (IArg)svcUuid, (IArg)paramID);
  user_enqueueCharDataMsg(APP_MSG_SERVICE_CFG, connHandle, svcUuid,
                          paramID, pValue, len);
}

{% endif -%}
/*
 *  Callbacks from Swi-context
 *****************************************************************************/

{% if pins.numButtons -%}
/*
 * @brief  Callback from Clock module on timeout
 *
 *         Determines new state after debouncing
 *
 * @param  buttonId    The pin being debounced
 */
static void buttonDebounceSwiFxn(UArg buttonId)
{
  // Used to send message to app
  button_state_t buttonMsg = { .pinId = buttonId };
  uint8_t        sendMsg   = FALSE;

  // Get current value of the button pin after the clock timeout
  uint8_t buttonPinVal = PIN_getInputValue(buttonId);

  // Set interrupt direction to opposite of debounced state
  // If button is now released (button is active low, so release is high)
  if (buttonPinVal)
  {
    // Enable negative edge interrupts to wait for press
    PIN_setConfig(userPinHandle, PIN_BM_IRQ, buttonId | PIN_IRQ_NEGEDGE);
  }
  else
  {
    // Enable positive edge interrupts to wait for relesae
    PIN_setConfig(userPinHandle, PIN_BM_IRQ, buttonId | PIN_IRQ_POSEDGE);
  }

 switch(buttonId)
 {
{%   for pin in pins.list -%}
{%     if pin.button -%}
    case {{ pin.define }}:
      // If button is now released (buttonPinVal is active low, so release is 1)
      // and button state was pressed (buttonstate is active high so press is 1)
      if (buttonPinVal && {{pin.nameCap}}State)
      {
        // Button was released
        buttonMsg.state = {{pin.nameCap}}State = 0;
        sendMsg = TRUE;
      }
      else if (!buttonPinVal && !{{pin.nameCap}}State)
      {
        // Button was pressed
        buttonMsg.state = {{pin.nameCap}}State = 1;
        sendMsg = TRUE;
      }
      break;
{%     endif -%}
{%   endfor -%}
  }

  if (sendMsg == TRUE)
  {
    user_enqueueRawAppMsg(APP_MSG_BUTTON_DEBOUNCED,
                      (uint8_t *)&buttonMsg, sizeof(buttonMsg));
  }
}

{% endif -%}
{% if anyCCCDs and numCCCDs > pins.numInput -%}
/*
 * Handles the Software Interrupt resulting from timeout of the clock object(s)
 * used to demonstrate notifications/indications.
 */
static void user_generic_clockSwiHandler(uint16_t svcUuid, uint16_t paramID)
{
  uint8_t notiData[20];
  uint16_t notiLen = sizeof notiData;

  Log_info2("(SWI) Generic clock handler: svc(0x%04x) paramID(%d)", (IArg)svcUuid, (IArg)paramID);
{% if 0 -%}

This is an attempt to use Liquid to determine whether a variable is needed.
It is not straightforward.

{% else -%}
{%   assign needsJunk = 0 -%}
{%   for service in services -%}
{%   for char in service.charsHaveCCCDs -%}
{%     unless char.isWritable -%}
{%       assign needsJunk = needsJunk | plus: 1 -%}
{%     endunless -%}
{%   endfor -%}
{%   endfor -%}
{%   if needsJunk -%}
  static uint8_t someCounter = 0;
{%   endif -%}
{% endif -%}

  // Get loopback data if char is writable, otherwise fill with junk
  switch(svcUuid) {
{% for service in services -%}
{% if service.numCCCDs -%}
    case {{ service.name | upcase }}_SERV_UUID:
      switch(paramID) {
{%     for char in service.charsHaveCCCDs -%}
        case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
{%       if char.isWritable -%}
          // Characteristic is writable, so echo back contents.
          {{ service.nameCap }}_GetParameter( paramID, &notiLen, notiData );
{%       else -%}
          // Characteristic is not writable, so send something, max 20 bytes (max default ATT MTU size minus header)
          notiLen = MIN(notiLen, {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_LEN);
          memset(notiData, someCounter++, notiLen);
{%       endif -%}
          break;
{%     endfor -%}
      }
      break;
{% endif -%}
{% endfor -%}
  }

  // Send message to application that it should update the value of the characteristic from Task context.
  user_enqueueCharDataMsg(APP_MSG_UPDATE_CHARVAL, 0xFFFF, svcUuid, paramID, notiData, notiLen);
}
{% endif -%}
{% for service in services -%}
{%   if service.numCCCDs and service.numCCCDs > service.pins.numInput -%}

/*
 * Swi handler for clock object(s) used by {{service.nameNice}}.
 * paramID is stored in the clock object for each characteristic.
 */
static void user_{{service.nameCap}}_clockSwiHandler(UArg paramID)
{
  // Act as a closure for the generic clockSwiHandler, as clock objects only have one parameter associated.
  user_generic_clockSwiHandler({{ service.name | upcase }}_SERV_UUID, paramID);
}
{%   endif -%}
{% endfor -%}

/*
 *  Callbacks from Hwi-context
 *****************************************************************************/
{% if pins.numInput -%}

/*
 * @brief  Callback from PIN driver on interrupt
 *
 *         Sets in motion the debouncing.
 *
 * @param  handle    The PIN_Handle instance this is about
 * @param  pinId     The pin that generated the interrupt
 */
static void pinHwiFxn(PIN_Handle handle, PIN_Id pinId)
{
  // Whether to disable Hwi for 'pinId' after handling the interrupt
  int disableHwi = 0;
  // Get current value of the button pin after the clock timeout
  uint8_t pinVal = PIN_getInputValue(pinId);

  Log_info1("Pin interrupt: IOID_%d", (IArg)pinId);

  switch (pinId)
  {
{%   for pin in pins.list -%}
{%     if pin.input or pin.button -%}
    case {{ pin.define }}:
{%       if pin.button -%}
      // Start debounce clock for this pin
      Clock_start(Clock_handle(&{{pin.nameCap}}DebounceClock));
      disableHwi = 1;  // Disable interrupt until debounce is complete
{%       else -%}
      // Send message to application that it should update the value of the characteristic from Task context.
      user_enqueueCharDataMsg(APP_MSG_UPDATE_CHARVAL, 0xFFFF,
                              {{ pin.service.name | upcase }}_SERV_UUID,
                              {{ pin.service.nameAbbr | upcase }}_{{ pin.char.name | upcase }}_ID,
                              &pinVal, 1);
{%       endif -%}
      break;
{%     endif -%}
{%   endfor -%}
  }

  if (disableHwi)
  {
    // Disable interrupt on that pin for now. Re-enabled after debounce.
    PIN_setConfig(handle, PIN_BM_IRQ, pinId | PIN_IRQ_DIS);
  }
}
{% endif -%}


/******************************************************************************
 *****************************************************************************
 *
 *  Utility functions
 *
 ****************************************************************************
 *****************************************************************************/
{% if anyWriteable or anyCCCDs -%}

/*
 * @brief  Generic message constructor for characteristic data.
 *
 *         Sends a message to the application for handling in Task context where
 *         the message payload is a char_data_t struct.
 *
 *         From service callbacks the appMsgType is APP_MSG_SERVICE_WRITE or
 *         APP_MSG_SERVICE_CFG, and functions running in another context than
 *         the Task itself, can set the type to APP_MSG_UPDATE_CHARVAL to
 *         make the user Task loop invoke user_updateCharVal function for them.
 *
 * @param  appMsgType    Enumerated type of message being sent.
 * @param  connHandle    GAP Connection handle of the relevant connection
 * @param  serviceUUID   16-bit part of the relevant service UUID
 * @param  paramID       Index of the characteristic in the service
 * @oaram  *pValue       Pointer to characteristic value
 * @param  len           Length of characteristic data
 */
static void user_enqueueCharDataMsg( app_msg_types_t appMsgType,
                                     uint16_t connHandle,
                                     uint16_t serviceUUID, uint8_t paramID,
                                     uint8_t *pValue, uint16_t len )
{
  // Called in Stack's Task context, so can't do processing here.
  // Send message to application message queue about received data.
  uint16_t readLen = len; // How much data was written to the attribute

  // Allocate memory for the message.
  // Note: The pCharData message doesn't have to contain the data itself, as
  //       that's stored in a variable in the service implementation.
  //
  //       However, to prevent data loss if a new value is received before the
  //       service's container is read out via the GetParameter API is called,
  //       we copy the characteristic's data now.
  app_msg_t *pMsg = ICall_malloc( sizeof(app_msg_t) + sizeof(char_data_t) +
                                  readLen );

  if (pMsg != NULL)
  {
    pMsg->type = appMsgType;

    char_data_t *pCharData = (char_data_t *)pMsg->pdu;
    pCharData->svcUUID = serviceUUID; // Use 16-bit part of UUID.
    pCharData->paramID = paramID;
    // Copy data from service now.
    memcpy(pCharData->data, pValue, readLen);
    // Update pCharData with how much data we received.
    pCharData->dataLen = readLen;
    // Enqueue the message using pointer to queue node element.
    Queue_enqueue(hApplicationMsgQ, &pMsg->_elem);
  // Let application know there's a message.
  Event_post(syncEvent, PRZ_APP_MSG_EVT);
  }
}

{% endif -%}
/*
 * @brief  Generic message constructor for application messages.
 *
 *         Sends a message to the application for handling in Task context.
 *
 * @param  appMsgType    Enumerated type of message being sent.
 * @oaram  *pValue       Pointer to characteristic value
 * @param  len           Length of characteristic data
 */
static void user_enqueueRawAppMsg(app_msg_types_t appMsgType, uint8_t *pData,
                                  uint16_t len)
{
  // Allocate memory for the message.
  app_msg_t *pMsg = ICall_malloc( sizeof(app_msg_t) + len );

  if (pMsg != NULL)
  {
    pMsg->type = appMsgType;

    // Copy data into message
    memcpy(pMsg->pdu, pData, len);

    // Enqueue the message using pointer to queue node element.

    Queue_enqueue(hApplicationMsgQ, &pMsg->_elem);
//    // Let application know there's a message.
    Event_post(syncEvent, PRZ_APP_MSG_EVT);
  }
}


/*
 * @brief  Convenience function for updating characteristic data via char_data_t
 *         structured message.
 *
 * @note   Must run in Task context in case BLE Stack APIs are invoked.
 *
 * @param  *pCharData  Pointer to struct with value to update.
 */
static void user_updateCharVal(char_data_t *pCharData)
{
  switch(pCharData->svcUUID)
  {
{% for service in services -%}
{%   if service.numCCCDs or service.numReadable -%}
    case {{ service.name | upcase }}_SERV_UUID:
      {{ service.nameCap }}_SetParameter( pCharData->paramID,
              pCharData->dataLen, pCharData->data );
    break;

{%   endif -%}
{% endfor -%}
  }
}

/*
 * @brief   Convert {0x01, 0x02} to "01:02"
 *
 * @param   src - source byte-array
 * @param   src_len - length of array
 * @param   dst - destination string-array
 * @param   dst_len - length of array
 *
 * @return  array as string
 */
static char *Util_convertArrayToHexString(uint8_t const *src, uint8_t src_len,
                                          uint8_t *dst, uint8_t dst_len)
{
  char        hex[] = "0123456789ABCDEF";
  uint8_t     *pStr = dst;
  uint8_t     avail = dst_len-1;

  memset(dst, 0, avail);

  while (src_len && avail > 3)
  {
    if (avail < dst_len-1) { *pStr++ = ':'; avail -= 1; };
    *pStr++ = hex[*src >> 4];
    *pStr++ = hex[*src++ & 0x0F];
    avail -= 2;
    src_len--;
  }

  if (src_len && avail)
    *pStr++ = ':'; // Indicate not all data fit on line.

  return (char *)dst;
}

/*
 * @brief   Extract the LOCALNAME from Scan/AdvData
 *
 * @param   data - Pointer to the advertisement or scan response data
 *
 * @return  Pointer to null-terminated string with the adv local name.
 */
static char *Util_getLocalNameStr(const uint8_t *data) {
  uint8_t nuggetLen = 0;
  uint8_t nuggetType = 0;
  uint8_t advIdx = 0;

  static char localNameStr[32] = { 0 };
  memset(localNameStr, 0, sizeof(localNameStr));

  for (advIdx = 0; advIdx < 32;) {
    nuggetLen = data[advIdx++];
    nuggetType = data[advIdx];
    if ( (nuggetType == GAP_ADTYPE_LOCAL_NAME_COMPLETE ||
          nuggetType == GAP_ADTYPE_LOCAL_NAME_SHORT) && nuggetLen < 31) {
      memcpy(localNameStr, &data[advIdx + 1], nuggetLen - 1);
      break;
    } else {
      advIdx += nuggetLen;
    }
  }

  return localNameStr;
}


/*********************************************************************
 *********************************************************************/

