{% if fluff -%}/******************************************************************************
 * Filename:       {{service.name}}.c
 *
 * Description:    This file contains the implementation of the service.
 *
 *                 Generated by:
 *                 BDS version: {{pluginInfo.AppVersion}}
 *                 Plugin:      {{pluginInfo.Name}} {{ pluginInfo.Version}}
 *                 Time:        {{pluginInfo.timestamp}}
 *

 * Copyright (c) 2015-2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 **********************************************************************************/


/*********************************************************************
 * INCLUDES
 */
#include <string.h>

//#define xdc_runtime_Log_DISABLE_ALL 1  // Add to disable logs from this file
#include <xdc/runtime/Log.h>
#include <xdc/runtime/Diags.h>
#ifdef UARTLOG_ENABLE
#  include "UartLog.h"
#endif

#include <icall.h>
#include "util.h"
/* This Header file contains all BLE API and icall structure definition */
#include "icall_ble_api.h"

#include "{{service.filename}}.h"

/*********************************************************************
 * MACROS
 */

/*********************************************************************
 * CONSTANTS
 */

/*********************************************************************
 * TYPEDEFS
 */

/*********************************************************************
* GLOBAL VARIABLES
*/

{% endif -%}// {{service.name}} Service UUID
CONST uint8_t {{service.nameCap}}UUID[{{ service.uuid_size }}] =
{
{% if service.uuid_size == 'ATT_UUID_SIZE' -%}
  {{ service.name | upcase }}_SERV_UUID_BASE128({{ service.name | upcase }}_SERV_UUID)
{% else -%}
  LO_UINT16({{ service.name | upcase }}_SERV_UUID), HI_UINT16({{ service.name | upcase }}_SERV_UUID)
{% endif -%}
};

{% for char in service.chars -%}
// {{char.name}} UUID
CONST uint8_t {{ service.nameAbbr}}_{{ char.nameCap }}UUID[{{ char.uuid_size }}] =
{
{% if char.uuid_size == 'ATT_UUID_SIZE' -%}
  {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_UUID_BASE128({{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_UUID)
{% else -%}
  LO_UINT16({{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_UUID), HI_UINT16({{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_UUID)
{% endif -%}
};

{% endfor -%}

{% if fluff -%}/*********************************************************************
 * LOCAL VARIABLES
 */

{% endif -%}
{% if service.numCCCDs or service.numWriteable -%}
static {{service.nameCap}}CBs_t *pAppCBs = NULL;
{% endif -%}
{% if service.numCCCDs -%}
static uint8_t {{ service.nameAbbr}}_icall_rsp_task_id = INVALID_TASK_ID;
{% endif -%}

{% if fluff -%}/*********************************************************************
* Profile Attributes - variables
*/

{% endif -%}// Service declaration
static CONST gattAttrType_t {{service.nameCap}}Decl = { {{ service.uuid_size }}, {{service.nameCap}}UUID };
{% for char in service.chars -%}

// Characteristic "{{ char.nameNice }}" Properties (for declaration)
static uint8_t {{ service.nameAbbr}}_{{ char.nameCap }}Props = {% if char.orProps %}{{ char.orProps }};{% else %}0;
#warning No properties provided. Did you remember to save the characteristic in the BDS GUI?{% endif %}

// Characteristic "{{ char.nameNice }}" Value variable
static uint8_t {{ service.nameAbbr}}_{{ char.nameCap }}Val[{{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN] = {0};
{% if service.numWriteable or service.numNeedRead -%}

// Length of data in characteristic "{{ char.nameNice }}" Value variable, initialized to minimal size.
static uint16_t {{ service.nameAbbr}}_{{ char.nameCap }}ValLen = {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN_MIN;
{% endif -%}
{% if char.hasCCCD -%}

// Characteristic "{{ char.nameNice }}" Client Characteristic Configuration Descriptor
static gattCharCfg_t *{{ service.nameAbbr}}_{{ char.nameCap }}Config;
{% endif -%}


{% endfor -%}

{% if fluff -%}/*********************************************************************
* Profile Attributes - Table
*/

{% endif -%}static gattAttribute_t {{service.name}}AttrTbl[] =
{
  // {{service.name}} Service Declaration
  {
    { ATT_BT_UUID_SIZE, primaryServiceUUID },
    GATT_PERMIT_READ,
    0,
    (uint8_t *)&{{service.nameCap}}Decl
  },
{% for char in service.chars -%}
    // {{ char.nameNice }} Characteristic Declaration
    {
      { ATT_BT_UUID_SIZE, characterUUID },
      GATT_PERMIT_READ,
      0,
      &{{ service.nameAbbr}}_{{ char.nameCap }}Props
    },
      // {{ char.nameNice }} Characteristic Value
      {
        { {{ char.uuid_size }}, {{ service.nameAbbr}}_{{ char.nameCap }}UUID },
        {% if char.orPerms %}{{ char.orPerms }}{% else %}0{% endif %},
        0,
        {{ service.nameAbbr}}_{{ char.nameCap }}Val
      },
{% if char.hasCCCD -%}
      // {{ char.nameNice }} CCCD
      {
        { ATT_BT_UUID_SIZE, clientCharCfgUUID },
        GATT_PERMIT_READ | GATT_PERMIT_WRITE,
        0,
        (uint8_t *)&{{ service.nameAbbr}}_{{ char.nameCap }}Config
      },
{% endif -%}
{% endfor -%}
};

{% if fluff -%}/*********************************************************************
 * LOCAL FUNCTIONS
 */
{% endif -%}static bStatus_t {{service.name}}_ReadAttrCB( uint16_t connHandle, gattAttribute_t *pAttr,
                                           uint8_t *pValue, uint16_t *pLen, uint16_t offset,
                                           uint16_t maxLen, uint8_t method );
static bStatus_t {{service.name}}_WriteAttrCB( uint16_t connHandle, gattAttribute_t *pAttr,
                                            uint8_t *pValue, uint16_t len, uint16_t offset,
                                            uint8_t method );

{% if fluff -%}/*********************************************************************
 * PROFILE CALLBACKS
 */
{% endif -%}// Simple Profile Service Callbacks
CONST gattServiceCBs_t {{service.name}}CBs =
{
  {{service.name}}_ReadAttrCB,  // Read callback function pointer
  {{service.name}}_WriteAttrCB, // Write callback function pointer
  NULL                       // Authorization callback function pointer
};

{% if fluff -%}/*********************************************************************
* PUBLIC FUNCTIONS
*/

/*
 * {{ service.nameCap }}_AddService- Initializes the {{ service.nameCap }} service by registering
 *          GATT attributes with the GATT server.
 *
 *    rspTaskId - The ICall Task Id that should receive responses for Indications.
 */
{% endif -%}extern bStatus_t {{ service.nameCap }}_AddService( uint8_t rspTaskId )
{
  uint8_t status;

{% for char in service.chars -%}
{% if char.hasCCCD -%}
  // Allocate Client Characteristic Configuration table
  {{ service.nameAbbr}}_{{ char.nameCap }}Config = (gattCharCfg_t *)ICall_malloc( sizeof(gattCharCfg_t) * linkDBNumConns );
  if ( {{ service.nameAbbr}}_{{ char.nameCap }}Config == NULL )
  {
    return ( bleMemAllocError );
  }

  // Initialize Client Characteristic Configuration attributes
  GATTServApp_InitCharCfg( INVALID_CONNHANDLE, {{ service.nameAbbr}}_{{ char.nameCap }}Config );
{% endif -%}
{% endfor -%}
  // Register GATT attribute list and CBs with GATT Server App
  status = GATTServApp_RegisterService( {{service.name}}AttrTbl,
                                        GATT_NUM_ATTRS( {{service.name}}AttrTbl ),
                                        GATT_MAX_ENCRYPT_KEY_SIZE,
                                        &{{service.name}}CBs );
  Log_info1("Registered service, %d attributes", (IArg)GATT_NUM_ATTRS( {{service.name}}AttrTbl ));
{% if service.numCCCDs -%}
  {{ service.nameAbbr}}_icall_rsp_task_id = rspTaskId;

{% endif -%}
  return ( status );
}

{% if fluff -%}/*
 * {{ service.nameCap }}_RegisterAppCBs - Registers the application callback function.
 *                    Only call this function once.
 *
 *    appCallbacks - pointer to application callbacks.
 */
{% endif -%}bStatus_t {{ service.nameCap }}_RegisterAppCBs( {{service.nameCap}}CBs_t *appCallbacks )
{
  if ( appCallbacks )
  {
{% if service.numCCCDs or service.numWriteable -%}
    pAppCBs = appCallbacks;
    Log_info1("Registered callbacks to application. Struct %p", (IArg)appCallbacks);
{% endif -%}
    return ( SUCCESS );
  }
  else
  {
    Log_warning0("Null pointer given for app callbacks.");
    return ( FAILURE );
  }
}

{% if fluff -%}/*
 * {{ service.nameCap }}_SetParameter - Set a {{ service.nameCap }} parameter.
 *
 *    param - Profile parameter ID
 *    len   - length of data to write
 *    value - pointer to data to write.  This is dependent on
 *            the parameter ID and may be cast to the appropriate
 *            data type (example: data type of uint16_t will be cast to
 *            uint16_t pointer).
 */
{% endif -%}
bStatus_t {{ service.nameCap }}_SetParameter( uint8_t param, uint16_t len, void *value )
{
{% if service.numCCCDs or service.numReadable -%}
  bStatus_t ret = SUCCESS;
  uint8_t  *pAttrVal;
  uint16_t *pValLen;
  uint16_t valMinLen;
  uint16_t valMaxLen;
{% endif -%}
{% if service.numCCCDs -%}
  uint8_t sendNotiInd = FALSE;
  gattCharCfg_t *attrConfig;
  uint8_t needAuth;
{% endif -%}

  switch ( param )
  {
{% for char in service.charsNeedRead -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
      pAttrVal  =  {{ service.nameAbbr}}_{{ char.nameCap }}Val;
      pValLen   = &{{ service.nameAbbr}}_{{ char.nameCap }}ValLen;
      valMinLen =  {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN_MIN;
      valMaxLen =  {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN;
{% if char.hasCCCD -%}
      sendNotiInd = TRUE;
      attrConfig  = {{ service.nameAbbr}}_{{ char.nameCap }}Config;
      needAuth    = FALSE; // Change if authenticated link is required for sending.
{% endif -%}
      Log_info2("SetParameter : %s len: %d", (IArg)"{{ char.nameCap }}", (IArg)len);
      break;

{% endfor -%}
    default:
      Log_error1("SetParameter: Parameter #%d not valid.", (IArg)param);
      return INVALIDPARAMETER;
  }
{% if service.numCCCDs or service.numReadable -%}

  // Check bounds, update value and send notification or indication if possible.
  if ( len <= valMaxLen && len >= valMinLen )
  {
    memcpy(pAttrVal, value, len);
    *pValLen = len; // Update length for read and get.
{% if service.numCCCDs -%}

    if (sendNotiInd)
    {
      Log_info2("Transmitting noti/ind- connHandle %d, %s", (IArg)attrConfig[0].connHandle,
                                                    (IArg)( (attrConfig[0].value==0)?"Noti/ind disabled":
                                                    (attrConfig[0].value==1)?"Notification enabled":
                                                     "Indication enabled" ) );
      // Try to send notification.
      GATTServApp_ProcessCharCfg( attrConfig, pAttrVal, needAuth,
                                  {{ service.name}}AttrTbl, GATT_NUM_ATTRS( {{ service.name}}AttrTbl ),
                                  {{ service.nameAbbr}}_icall_rsp_task_id,  {{service.name}}_ReadAttrCB);
    }
{% endif -%}
  }
  else
  {
    Log_error3("Length outside bounds: Len: %d MinLen: %d MaxLen: %d.", (IArg)len, (IArg)valMinLen, (IArg)valMaxLen);
    ret = bleInvalidRange;
  }

  return ret;
{% endif -%}
}


{% if fluff -%}/*
 * {{ service.nameCap }}_GetParameter - Get a {{ service.nameCap }} parameter.
 *
 *    param - Profile parameter ID
 *    len   - pointer to a variable that contains the maximum length that can be written to *value.
              After the call, this value will contain the actual returned length.
 *    value - pointer to data to write.  This is dependent on
 *            the parameter ID and may be cast to the appropriate
 *            data type (example: data type of uint16_t will be cast to
 *            uint16_t pointer).
 */
{% endif -%}bStatus_t {{ service.nameCap }}_GetParameter( uint8_t param, uint16_t *len, void *value )
{
  bStatus_t ret = SUCCESS;
  switch ( param )
  {
{% for char in service.charsNeedWrite -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
      *len = MIN(*len, {{ service.nameAbbr}}_{{ char.nameCap }}ValLen);
      memcpy(value, {{ service.nameAbbr}}_{{ char.nameCap }}Val, *len);
      Log_info2("GetParameter : %s returning %d bytes", (IArg)"{{ char.nameCap }}", (IArg)*len);
      break;

{% endfor -%}
    default:
      Log_error1("GetParameter: Parameter #%d not valid.", (IArg)param);
      ret = INVALIDPARAMETER;
      break;
  }
  return ret;
}

{% if fluff -%}/*********************************************************************
 * @internal
 * @fn          {{service.name}}_findCharParamId
 *
 * @brief       Find the logical param id of an attribute in the service's attr table.
 *
 *              Works only for Characteristic Value attributes and
 *              Client Characteristic Configuration Descriptor attributes.
 *
 * @param       pAttr - pointer to attribute
 *
 * @return      uint8_t paramID (ref {{service.name}}.h) or 0xFF if not found.
 */
{% endif -%}static uint8_t {{service.name}}_findCharParamId(gattAttribute_t *pAttr)
{
  // Is this a Client Characteristic Configuration Descriptor?
  if (ATT_BT_UUID_SIZE == pAttr->type.len && GATT_CLIENT_CHAR_CFG_UUID == *(uint16_t *)pAttr->type.uuid)
    return {{service.name}}_findCharParamId(pAttr - 1); // Assume the value attribute precedes CCCD and recurse

{% for char in service.chars -%}
  // Is this attribute in "{{ char.nameNice }}"?
  else if ( {{char.uuid_size}} == pAttr->type.len && !memcmp(pAttr->type.uuid, {{ service.nameAbbr}}_{{ char.nameCap }}UUID, pAttr->type.len))
    return {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_ID;

{% endfor -%}
  else
    return 0xFF; // Not found. Return invalid.
}

{% if fluff -%}/*********************************************************************
 * @fn          {{service.name}}_ReadAttrCB
 *
 * @brief       Read an attribute.
 *
 * @param       connHandle - connection message was received on
 * @param       pAttr - pointer to attribute
 * @param       pValue - pointer to data to be read
 * @param       pLen - length of data to be read
 * @param       offset - offset of the first octet to be read
 * @param       maxLen - maximum length of data to be read
 * @param       method - type of read message
 *
 * @return      SUCCESS, blePending or Failure
 */
{% endif -%}static bStatus_t {{service.name}}_ReadAttrCB( uint16_t connHandle, gattAttribute_t *pAttr,
                                       uint8_t *pValue, uint16_t *pLen, uint16_t offset,
                                       uint16_t maxLen, uint8_t method )
{
{% if service.numReadable or service.numCCCDs -%}
  bStatus_t status = SUCCESS;
  uint16_t valueLen;
{% endif -%}
  uint8_t paramID = 0xFF;

  // Find settings for the characteristic to be read.
  paramID = {{service.name}}_findCharParamId( pAttr );
  switch ( paramID )
  {
{% for char in service.charsNeedRead -%}
    case {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_ID:
      valueLen = {{ service.nameAbbr}}_{{ char.nameCap }}ValLen;

      Log_info4("ReadAttrCB : %s connHandle: %d offset: %d method: 0x%02x",
                 (IArg)"{{ char.nameCap }}",
                 (IArg)connHandle,
                 (IArg)offset,
                 (IArg)method);
      /* Other considerations for {{ char.nameNice }} can be inserted here */
      break;

{% endfor -%}
    default:
      Log_error0("Attribute was not found.");
      return ATT_ERR_ATTR_NOT_FOUND;
  }
{% if service.numReadable or service.numCCCDs -%}
  // Check bounds and return the value
  if ( offset > valueLen )  // Prevent malicious ATT ReadBlob offsets.
  {
    Log_error0("An invalid offset was requested.");
    status = ATT_ERR_INVALID_OFFSET;
  }
  else
  {
    *pLen = MIN(maxLen, valueLen - offset);  // Transmit as much as possible
    memcpy(pValue, pAttr->pValue + offset, *pLen);
  }

  return status;
{% endif -%}
}

{% if fluff -%}/*********************************************************************
 * @fn      {{service.name}}_WriteAttrCB
 *
 * @brief   Validate attribute data prior to a write operation
 *
 * @param   connHandle - connection message was received on
 * @param   pAttr - pointer to attribute
 * @param   pValue - pointer to data to be written
 * @param   len - length of data
 * @param   offset - offset of the first octet to be written
 * @param   method - type of write message
 *
 * @return  SUCCESS, blePending or Failure
 */
{% endif -%}static bStatus_t {{service.name}}_WriteAttrCB( uint16_t connHandle, gattAttribute_t *pAttr,
                                        uint8_t *pValue, uint16_t len, uint16_t offset,
                                        uint8_t method )
{
{% if service.numWriteable  or service.numCCCDs -%}
  bStatus_t status  = SUCCESS;
{% endif -%}
  uint8_t   paramID = 0xFF;
{% if service.numWriteable -%}
  uint8_t   changeParamID = 0xFF;
  uint16_t writeLenMin;
  uint16_t writeLenMax;
  uint16_t *pValueLenVar;
{% endif -%}
{% if service.numCCCDs -%}

  // See if request is regarding a Client Characterisic Configuration
  if (ATT_BT_UUID_SIZE == pAttr->type.len && GATT_CLIENT_CHAR_CFG_UUID == *(uint16_t *)pAttr->type.uuid)
  {
    // Allow notification and indication, but do not check if really allowed per CCCD.
    status = GATTServApp_ProcessCCCWriteReq( connHandle, pAttr, pValue, len,
                                             offset, GATT_CLIENT_CFG_NOTIFY |
                                                     GATT_CLIENT_CFG_INDICATE );
    if (SUCCESS == status && pAppCBs && pAppCBs->pfnCfgChangeCb)
       pAppCBs->pfnCfgChangeCb( connHandle, {{ service.name | upcase }}_SERV_UUID,
                                {{service.name}}_findCharParamId(pAttr), pValue, len );

     return status;
  }
{% endif -%}

  // Find settings for the characteristic to be written.
  paramID = {{service.name}}_findCharParamId( pAttr );
  switch ( paramID )
  {
{% for char in service.charsNeedWrite -%}
    case {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_ID:
      writeLenMin  = {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN_MIN;
      writeLenMax  = {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_LEN;
      pValueLenVar = &{{ service.nameAbbr}}_{{ char.nameCap }}ValLen;

      Log_info5("WriteAttrCB : %s connHandle(%d) len(%d) offset(%d) method(0x%02x)",
                 (IArg)"{{ char.nameCap }}",
                 (IArg)connHandle,
                 (IArg)len,
                 (IArg)offset,
                 (IArg)method);
      /* Other considerations for {{ char.nameNice }} can be inserted here */
      break;

{% endfor -%}
    default:
      Log_error0("Attribute was not found.");
      return ATT_ERR_ATTR_NOT_FOUND;
  }
{% if service.numWriteable -%}
  // Check whether the length is within bounds.
  if ( offset >= writeLenMax )
  {
    Log_error0("An invalid offset was requested.");
    status = ATT_ERR_INVALID_OFFSET;
  }
  else if ( offset + len > writeLenMax )
  {
    Log_error0("Invalid value length was received.");
    status = ATT_ERR_INVALID_VALUE_SIZE;
  }
  else if ( offset + len < writeLenMin && ( method == ATT_EXECUTE_WRITE_REQ || method == ATT_WRITE_REQ ) )
  {
    // Refuse writes that are lower than minimum.
    // Note: Cannot determine if a Reliable Write (to several chars) is finished, so those will
    //       only be refused if this attribute is the last in the queue (method is execute).
    //       Otherwise, reliable writes are accepted and parsed piecemeal.
    Log_error0("Invalid value length was received.");
    status = ATT_ERR_INVALID_VALUE_SIZE;
  }
  else
  {
    // Copy pValue into the variable we point to from the attribute table.
    memcpy(pAttr->pValue + offset, pValue, len);

    // Only notify application and update length if enough data is written.
    //
    // Note: If reliable writes are used (meaning several attributes are written to using ATT PrepareWrite),
    //       the application will get a callback for every write with an offset + len larger than _LEN_MIN.
    // Note: For Long Writes (ATT Prepare + Execute towards only one attribute) only one callback will be issued,
    //       because the write fragments are concatenated before being sent here.
    if ( offset + len >= writeLenMin )
    {
      changeParamID = paramID;
      *pValueLenVar = offset + len; // Update data length.
    }
  }

  // Let the application know something changed (if it did) by using the
  // callback it registered earlier (if it did).
  if (changeParamID != 0xFF)
    if ( pAppCBs && pAppCBs->pfnChangeCb )
      pAppCBs->pfnChangeCb( connHandle, {{ service.name | upcase }}_SERV_UUID, paramID, pValue, len+offset ); // Call app function from stack task context.

  return status;
{% endif -%}
}
