{% if fluff -%}/******************************************************************************
 * Filename:       {{service.name}}.h
 *
 * Description:    This file contains the {{service.name}} service definitions and
 *                 prototypes.
 *
 *                 Generated by:
 *                 BDS version: {{pluginInfo.AppVersion}}
 *                 Plugin:      {{pluginInfo.Name}} {{pluginInfo.Version}}
 *                 Time:        {{pluginInfo.timestamp}}
 *

 * Copyright (c) 2015-2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

#ifndef _{{ service.filename | upcase }}_H_
#define _{{ service.filename | upcase }}_H_

#ifdef __cplusplus
extern "C"
{
#endif

/*********************************************************************
 * INCLUDES
 */
#include <bcomdef.h>

/*********************************************************************
 * CONSTANTS
 */
{% endif -%}// Service UUID
#define {{ service.name | upcase }}_SERV_UUID {{service.uuid}}
{% if service.uuid_size == 'ATT_UUID_SIZE' -%}
#define {{ service.name | upcase }}_SERV_UUID_BASE128(uuid) {{ service.baseUuidMacro }}
{% endif -%}

{% for char in service.chars -%}
// {{char.nameNice}} Characteristic defines
#define {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID                 {{ char.idx }}
#define {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_UUID               {{ char.uuid }}
{% if char.uuid_size == 'ATT_UUID_SIZE' -%}
#define {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_UUID_BASE128(uuid) {{ char.baseUuidMacro }}
{% endif -%}
#define {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_LEN                {{ char.len }}
#define {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_LEN_MIN            {{ char.lenMin }}

{% endfor -%}
{% if fluff -%}
/*********************************************************************
 * TYPEDEFS
 */
{% for char in service.chars -%}

// Fields in characteristic "{{ char.nameNice }}"
{%   for field in char.fields -%}
//   Field "{{field.name}}" format: {{field.format}}, bits: {{field.numBits}}
{%     for enum in field.enumerations -%}
{%       if forloop.first -%}
//     Enumerations for {{field.name}}
{%       endif -%}
//       {{enum.value}} -- {{enum.key}}
{%       if forloop.last -%}
{%comment%} Intentionally left blank {%endcomment%}
{%       endif -%}
{%     endfor -%}
{%     for bit in field.bits -%}
{%       if forloop.first -%}
//     Bits in field "{{field.name}}"
{%       endif -%}
//       {{bit.name}} -- Offset: {{bit.index}}, Size = {{bit.size}}
{%       for enum in bit.enumerations -%}
{%         if forloop.first -%}
//         Enumerations for {{bit.name}}
{%         endif -%}
//           {{enum.value}} -- {{enum.key}}
{%         if forloop.last -%}
{%comment%} Intentionally left blank {%endcomment%}
{%         endif -%}
{%       endfor -%}
{%     endfor -%}
{%   endfor -%}
{% endfor -%}

/*********************************************************************
 * MACROS
 */

/*********************************************************************
 * Profile Callbacks
 */

{% endif -%}// Callback when a characteristic value has changed
typedef void (*{{service.nameCap}}Change_t)( uint16_t connHandle, uint16_t svcUuid, uint8_t paramID, uint8_t *pValue, uint16_t len );

typedef struct
{
  {{service.nameCap}}Change_t        pfnChangeCb;     // Called when characteristic value changes
  {{service.nameCap}}Change_t        pfnCfgChangeCb;  // Called when characteristic CCCD changes
} {{service.nameCap}}CBs_t;

{% if fluff -%}


/*********************************************************************
 * API FUNCTIONS
 */


/*
 * {{ service.nameCap }}_AddService- Initializes the {{ service.nameCap }} service by registering
 *          GATT attributes with the GATT server.
 *
 *    rspTaskId - The ICall Task Id that should receive responses for Indications.
 */
{% endif -%}extern bStatus_t {{ service.nameCap }}_AddService( uint8_t rspTaskId );

{% if fluff -%}/*
 * {{ service.nameCap }}_RegisterAppCBs - Registers the application callback function.
 *                    Only call this function once.
 *
 *    appCallbacks - pointer to application callbacks.
 */
{% endif -%}extern bStatus_t {{ service.nameCap }}_RegisterAppCBs( {{service.nameCap}}CBs_t *appCallbacks );

{% if fluff -%}/*
 * {{ service.nameCap }}_SetParameter - Set a {{ service.nameCap }} parameter.
 *
 *    param - Profile parameter ID
 *    len   - length of data to write
 *    value - pointer to data to write.  This is dependent on
 *            the parameter ID and may be cast to the appropriate
 *            data type (example: data type of uint16_t will be cast to
 *            uint16_t pointer).
 */
{% endif -%}extern bStatus_t {{ service.nameCap }}_SetParameter( uint8_t param, uint16_t len, void *value );

{% if fluff -%}/*
 * {{ service.nameCap }}_GetParameter - Get a {{ service.nameCap }} parameter.
 *
 *    param - Profile parameter ID
 *    len   - pointer to a variable that contains the maximum length that can be written to *value.
              After the call, this value will contain the actual returned length.
 *    value - pointer to data to write.  This is dependent on
 *            the parameter ID and may be cast to the appropriate
 *            data type (example: data type of uint16_t will be cast to
 *            uint16_t pointer).
 */
{% endif -%}extern bStatus_t {{ service.nameCap }}_GetParameter( uint8_t param, uint16_t *len, void *value );
{% if fluff -%}

/*********************************************************************
*********************************************************************/

#ifdef __cplusplus
}
#endif

#endif /* _{{ service.name | upcase }}_H_ */
{% endif %}
